
#include "auth.h"
#include "debug.h"
#include "stringtools.h"
#include "domain_name_cache.h"
#include "xmalloc.h"

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

struct auth_ops {
	char *type;
	auth_assert_t assert;
	auth_accept_t accept;
	struct auth_ops *next;
};

static struct auth_ops *list=0;

static struct auth_ops * type_lookup( char *type )
{
	struct auth_ops *a;

	for(a=list;a;a=a->next) {
		if(!strcmp(a->type,type)) return a;
	}

	return 0;
}

static int auth_lookup( const char *authfile, const char *type, const char *host, const char *subject, char **local_user )
{
	char line[AUTH_LINE_MAX];
	char ntype[AUTH_LINE_MAX];
	char nhost[AUTH_LINE_MAX];
	char nsubject[AUTH_LINE_MAX];
	char lsubject[AUTH_LINE_MAX];
	int fields;
	int result=0;
	FILE *file;
	int linenum=0;

	file = fopen(authfile,"r");
	if(!file) return 0;

	while(fgets(line,sizeof(line),file)) {

		linenum++;
		string_chomp(line);

		if( line[0]=='#' ) continue;

		fields = sscanf(line,"%[^:]:%[^:]:%[^:]:%[^:]",ntype,nhost,nsubject,lsubject);
		if(fields==4) {
			if( string_match(ntype,type) &&
			    string_match(nhost,host) &&
			    string_match(nsubject,subject) ) {
				if(!strcmp(lsubject,"*")) {
					*local_user = xstrdup(subject);
				} else {
					*local_user = xstrdup(lsubject);
				}
				result = 1;
				break;
			}
		} else {
			debug("auth_lookup: bad data on line %d in file %s",linenum,authfile);
		}
	}

	fclose(file);
	return result;
}

int auth_assert( struct link *link, char **type, char **subject, time_t stoptime )
{
	char line[AUTH_LINE_MAX];
	struct auth_ops *a;

	for( a=list; a; a=a->next ) {

		debug("auth: requesting '%s' authentication",a->type);

		sprintf(line,"%s\n",a->type);
		link_write(link,line,strlen(line),stoptime);

		if(!link_readline(link,line,AUTH_LINE_MAX,stoptime)) break;
		if(!strcmp(line,"yes")) {
			debug("auth: server agrees to try '%s'",a->type);
			if(a->assert(link,stoptime)) {
				debug("auth: successfully authenticated");
				if(!link_readline(link,line,AUTH_LINE_MAX,stoptime)) break;
				if(!strcmp(line,"yes")) {
					debug("auth: authorized");
					debug("auth: reading back auth info from server");
					if(!link_readline(link,line,sizeof(line),stoptime)) return 0;
					*type = strdup(line);
					if(!link_readline(link,line,sizeof(line),stoptime)) return 0;
					*subject = strdup(line);
					return 1;
				} else {
					debug("auth: not authorized");
				}
			} else {
				debug("auth: failed to authenticate");
			}
		} else {
			debug("auth: server refuses to try '%s'",a->type);
		}
		debug("auth: still trying...");
	}

	if(!a) {
		debug("auth: ran out of authenticators");
	} else {
		debug("auth: lost connection");
	}

	return 0;
}

int auth_accept( struct link *link, const char *authfile, char **typeout, char **hostout, char **subject, char **local_user, time_t stoptime )
{
	struct auth_ops *a;
	char type[AUTH_TYPE_MAX];
	char line[AUTH_LINE_MAX];
	char addr[LINK_ADDRESS_MAX];
	char name[DOMAIN_NAME_MAX];
	int port;

	*subject = 0;

	link_address_remote(link,addr,&port);
	if(domain_name_cache_lookup_reverse(addr,name)) {
		debug("auth: client %s is %s",addr,name);
	} else {
		debug("auth: client %s has no domain name",addr);
		strcpy(name,addr);
	}

	while(link_readline(link,type,AUTH_TYPE_MAX,stoptime)) {

		string_chomp(line);

		debug("auth: client requests '%s' authentication",type);

		a = type_lookup( type );
		if(a) {
			debug("auth: I agree to try '%s' ",type);
			link_write(link,"yes\n",4,stoptime);
		} else {
			debug("auth: I do not agree to '%s' ",type);
			link_write(link,"no\n",3,stoptime);
			continue;
		}

		if(a->accept(link,subject,stoptime)) {
			debug("auth: subject authenticated as %s",*subject);
			if(auth_lookup(authfile,type,name,*subject,local_user)) {
				debug("auth: subject is authorized");
				sprintf(line,"yes\n%s\n%s\n",type,*subject);
				link_write(link,line,strlen(line),stoptime);
				*typeout = xstrdup(type);
				*hostout = xstrdup(name);
				return 1;
			} else {
				free(*subject);
				*subject = 0;
				debug("auth: subject is not authorized");
				link_write(link,"no\n",3,stoptime);
			}
		} else {
			debug("auth: client could not authenticate using '%s'",type);
		}
		debug("auth: still trying");
	}

	debug("auth: client disconnected");

	return 0;
}


int auth_barrier( struct link *link, const char *response, time_t stoptime )
{
	char line[AUTH_LINE_MAX];

	link_write(link,response,strlen(response),stoptime);

	if(link_readline(link,line,sizeof(line),stoptime)) {
		if(!strcmp(line,"yes")) {
			return 1;
		}
	}

	return 0;
}

int auth_register( char *type, auth_assert_t assert, auth_accept_t accept )
{
	struct auth_ops *a;

	a = (struct auth_ops *) malloc(sizeof(struct auth_ops));
	if(!a) return 0;

	a->type = type;
	a->assert = assert;
	a->accept = accept;
	a->next = list;
	list = a;

	return 1;
}

