/*
dttools is Copyright (C) 2002 Douglas Thain
This software is released under the GNU General Public License.
See the file COPYING for details.
*/


#include "config_table.h"
#include "hash_table.h"
#include "username.h"
#include "domain_name_cache.h"
#include "stringtools.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

static char * expand_string( struct hash_table *hash, char *value );
static char * lookup_string( struct hash_table *hash, const char *value );

struct config_table {
	struct hash_table *hash;
};

struct config_table * config_table_create() {
	struct config_table *t;

	t = malloc(sizeof(*t));
	if(!t) return 0;


	t->hash = hash_table_create(127,hash_string);
	if(!t->hash) {
		free(t);
		return 0;
	}

	return t;
}

static int delete( struct hash_table *h, char *key, void *item, void *arg )
{
	free(item);
	return 1;
}

void config_table_delete( struct config_table *t )
{
	if(t) {
		if(t->hash) {
			hash_table_iterate(t->hash,delete,0);
			hash_table_delete(t->hash);
		}
		free(t);
	}
}

static int isdelim( char ch )
{
	switch(ch) {
		case ' ':
		case '\t':
		case ':':
		case '=':
			return 1;
		default:
			return 0;
	}
}

int    config_table_load( struct config_table *t, const char *filename )
{
	char line[CONFIG_TABLE_LINE_MAX];
	char *key, *value;
	FILE *file;
	int save_errno;

	file = fopen(filename,"r");
	if(!file) return 0;

	while(fgets(line,CONFIG_TABLE_LINE_MAX,file)) {
		string_chomp(line);
		if(line[0]=='#') continue;
		if(line[0]==0) continue;
		if(isdelim(line[0])) continue;

		key = &line[0];

		value = key;
		while(*value && !isdelim(*value)) value++;
		*value = 0;
		value++;
		while(isdelim(*value)) value++;
		if(!*value) continue;

		if(!strcmp(key,"include")) {
			value = strdup(value);
			if(!value) goto failure;

			value = expand_string(t->hash,value);
			if(!value) goto failure;

			if(!config_table_load(t,value)) goto failure;

			free(value);

		} else {
			if(!config_table_set(t,key,value)) goto failure;
		}
	}

	fclose(file);

	return 1;

	failure:
	save_errno = errno;
	if(file) fclose(file);
	errno = save_errno;
	return 0;
}

int config_table_set( struct config_table *t, const char *key, const char *value )
{
	void *old;
	char *copy;

	if(value) {
		copy = strdup(value);
		if(!copy) return 0;

		if(!hash_table_insert(t->hash,key,copy,&old)) {
			free(copy);
			return 0;
		}
	} else {
		old = hash_table_remove(t->hash,key);
	}

	if(old) free(old);


	return 1;
}

static char * get_special( const char *key )
{
	if(!strcmp(key,"USER")) {
		char name[USERNAME_MAX];
		if(username_get(name)) {
			return strdup(name);
		} else {
			return 0;
		}
	} else if( !strcmp(key,"HOST") || !strcmp(key,"HOSTNAME") ) {
		char name[DOMAIN_NAME_MAX];
		char *dot;

		if(domain_name_cache_guess(name)) {
			dot = strchr(name,'.');
			if(dot) *dot=0;
			return strdup(name);
		} else {
			return 0;
		}
	} else if( !strcmp(key,"FULLHOSTNAME") ) {
		char name[DOMAIN_NAME_MAX];
		if(domain_name_cache_guess(name)) {
			return strdup(name);
		} else {
			return 0;
		}
	} else {
		return 0;
	}
}

/*
Given a value, possible containing ${x} references,
expand the string and return the evaluated string.
*/

static char * expand_string( struct hash_table *hash, char *value )
{
	char *subvalue, *newvalue;
	char *dollar, *lbrace, *rbrace;
	int length;

	while(1) {
		dollar = strchr(value,'$');
		if(!dollar) return value;

		lbrace = strchr(dollar,'{');
		if(!lbrace) return value;

		rbrace = strchr(lbrace,'}');
		if(!rbrace) return value;

		*rbrace = 0;

		subvalue = lookup_string(hash,lbrace+1);
		if(!subvalue) return value;

		*rbrace = '}';

		length = strlen(value) - (rbrace-dollar) + strlen(subvalue) + 1;
		newvalue = malloc(length);
		if(!newvalue) {
			free(subvalue);
			free(value);
			return 0;
		}

		*dollar = 0;

		strcpy(newvalue,value);
		strcat(newvalue,subvalue);
		strcat(newvalue,rbrace+1);

		free(subvalue);
		free(value);

		value = newvalue;
	}
}


/*
Return the value of this key wrt this table.
Expand any subvalues as necessary.
*/

static char * lookup_string( struct hash_table *hash, const char *key )
{
	char *value;

	value = get_special(key);
	if(!value) {
		value = (char*) hash_table_lookup(hash,key);
		if(!value) return 0;

		value = strdup(value);
		if(!value) return 0;
	}

	if(value) {
		return expand_string(hash,value);
	} else {
		return 0;
	}
}

int config_table_get( struct config_table *t, const char *key, char **result )
{
	*result = lookup_string(t->hash,key);
	if(*result) {
		return 1;
	} else {
		return 0;
	}
}

int config_table_get_int( struct config_table *t, const char *key, INT64_T *value )
{
	char *s;
	if(config_table_get(t,key,&s)) {
		if(sscanf(s,INT64_FORMAT,value)) {
			free(s);
			return 1;
		} else {
			free(s);
			return 0;
		}
	} else {
		return 0;
	}
}

static int output( struct hash_table *hash, char *key, void *item, void *arg )
{
	char *value = lookup_string(hash,key);	
	if(value) {
		fprintf(arg,"%s=%s\n",key,value);
		free(value);
	} else {
		fprintf(arg,"%s=\n",key);
	}
	return 1;
}

int config_table_save( struct config_table *t, const char *filename )
{
	FILE *file;

	file = fopen(filename,"w");
	if(!file) return 0;

	if(!hash_table_iterate(t->hash,output,file)) {
		fclose(file);
		return 0;
	}
	
	fclose(file);
	return 1;
}


